from re      import split
from time    import time

from .common    import node_debug_message, is_string
from .exception import *
from .graph     import Vertex
from .kami      import Kami
from .selector  import PathType
from .parser    import selector as parse_selector

__all__ = ['Kotoba']

class Kotoba(Vertex):
    """
    XML Parser with Level-3 CSS Selectors

    :param kotoba.driver.IDriver node:

    Currently supported selectors:

    * selectors with at least one of element name, attributes and pseudo classes (only ``:root`` and ``:empty``)
    * all four combinations of selectors are supported (e.g., ``selector_1 operator_1 selector_2 ...``)
    * support wildcard search (only for element name)
    * support multi combinations in a single statement (e.g., ``combo_1, combo_2``)
    """

    static_guid = 1
    debug_mode  = False

    def __init__(self, node=None, level=0):
        """ Construct an XML parser using CSS3 selectors """
        super(self.__class__, self).__init__()

        self._guid   = Kotoba.static_guid
        self._level  = level
        self._node   = node
        self._parent = None
        self._data   = None
        self._adjacents  = Kami()
        self._children   = Kami()
        self._attributes = None

        # lazy-loading flag
        self._is_children_initialized = False

        self.name(node.name())

        # Take care of static data
        Kotoba.static_guid += 1

    def guid(self):
        return self._guid

    def level(self):
        return self._level

    def parent(self, parent=None):
        if parent:
            self._parent = parent

        return self._parent

    def attribute(self, key):
        if not key in self.attributes():
            return None

        return self.attributes()[key]

    def has_attribute(self, key):
        return key in self.attributes()

    def attributes(self):
        if not self._attributes:
            self._attributes = dict(self.node().attributes())

        return self._attributes

    def children(self, selector=None, include_data_blocks=False):
        if not self._is_children_initialized:
            self._node.__class__.initialize_children(self)

            self._is_children_initialized = True

        if is_string(selector):
            selector = parse_selector(selector)

        returnees = self._children

        if selector:
            returnees = self._find_descendants(selector, True)
        if include_data_blocks:
            returnees = self.adjacents()

        return returnees

    def find(self, selector):
        if is_string(selector):
            selector = parse_selector(selector)

        if not selector:
            raise InvalidSelectorError()

        returnees = Kotoba._direct_search(self, selector)

        return returnees

    def __repr__(self):
        return '<{}:{}>'.format(self.__class__.__name__, self.name())

    @staticmethod
    def _direct_search(node, selector):
        node_debug_message(node, node)

        search_type = selector.kind()

        if search_type == PathType.any_siblings or search_type == PathType.immediate_siblings:
            raise Runtime('Future Features')

        node_debug_message(node, '[SEARCH] DESCENDANTS/CHILDREN')

        # Continue the search from the descendants.
        return node._find_descendants(selector, search_type == PathType.children)

    def debug_message(self, message, ignore_indentation=False):
        node_debug_message(self, message, ignore_indentation)

    def _find_descendants(self, selector, stop_here=False):
        returnees = Kami()

        for child in self.children():
            if not child.is_element():
                raise RuntimeException('This child is not an element.')

            if selector.match(child): # if matched
                if not selector.next(): # if this is the last selector
                    returnees.append(child)
                    if not child.children():
                        continue

                    if not stop_here:
                        returnees.extend(child.find(selector))

                    continue

                if not stop_here:
                    returnees.extend(child.find(selector.next()))

                continue

            if not stop_here:
                returnees.extend(child.find(selector))

        return returnees

    def node(self):
        return self._node

    def original_value(self):
        """ Retrieve the value of the node (node driver).

            .. versionadded:: 3.1
        """

        return self._node.value()

    def kind(self):
        return self._node.nodeType

    def is_element(self):
        return self._node.is_element()

    def is_comment(self):
        return self._node.is_comment()

    def is_data(self):
        return self._node.is_data()

    def data(self):
        if not self._data:
            self._data = self._node.__class__.retrieve_data(self)

        return self._data

    def dump(self, label, force_print=False, ignore_indentation=False):
        if not force_print:
            return

        if label:
            self.debug_message('[{}]'.format(label), ignore_indentation)

        self.debug_message('CURRENT: {}'.format(self), ignore_indentation)
        self.debug_message('PARENT:  {}'.format(self.parent()), ignore_indentation)
        self.debug_message('KIND: {}'.format(self.kind()), ignore_indentation)

    def __str__(self):
        representative = 'NODE {} AT LEVEL {} ({})'.format(self.name(), self.level(), self._guid)

        if self.is_data():
            representative = 'NODE [DATA] AT LEVEL {} ({})'.format(self.level(), self._guid)
        elif not self.name():
            representative = 'NODE [TYPE-{}] AT LEVEL {} ({})'.format(self.kind(), self.level(), self._guid)

        return representative



import os
import sys
import unittest
import bootstrap

file_path = os.path.abspath(__file__)
code_path = os.path.dirname(os.path.abspath(file_path + '/../'))
test_path = os.path.dirname(file_path)

if code_path not in sys.path:
    sys.path.remove(code_path)

sys.path.insert(0, code_path)

if test_path not in sys.path:
    sys.path.remove(test_path)

sys.path.insert(0, test_path)

# suite = unittest.TestLoader().discover(
#     bootstrap.testing_base_path,
#     pattern='test_*.py'
# )
# unittest.TextTestRunner(verbosity=1).run(suite)



from .common    import is_string
from .exception import *

class Vertex(object):
    '''
    Data structure representing a vertex in a graph.
        
    *name* (required, string) is the name of the vertex.
        
    *adjacents* (optional, list, default: None) is a list of neighbours (*graph.Vertex*).
    '''
    
    def __init__(self):
        self._name      = None
        self._adjacents = []
    
    def name(self, new_name=None):
        if new_name and not is_string(new_name):
            raise InvalidInputError
        
        if is_string(new_name):
            self._name = new_name
        
        return self._name
    
    def adjacents(self):
        return self._adjacents


class InvalidDataSourceError(Exception):
    ''' Exception: The data source invalid '''

class InvalidSelectorError(Exception):
    ''' Exception: The selector invalid '''

class InvalidInputError(Exception):
    ''' Exception: Invalid input '''

class FileNotFoundError(Exception):
    ''' Exception: File not found '''

class LexicalError(Exception):
    ''' Exception: Selector Syntax Error '''


class IDriver(object):
    def name(self):
        raise NotImplementedError('Interface method not implemented')

    def attributes(self):
        raise NotImplementedError('Interface method not implemented')

    def children(self):
        raise NotImplementedError('Interface method not implemented')

    def value(self):
        raise NotImplementedError('Interface method not implemented')

    def is_element(self):
        raise NotImplementedError('Interface method not implemented')

    def is_comment(self):
        raise NotImplementedError('Interface method not implemented')

    def is_data(self):
        raise NotImplementedError('Interface method not implemented')

    @staticmethod
    def initialize_children(kotoba_node):
        raise NotImplementedError('Interface method not implemented')

class JSONDriver(IDriver):
    def __init__(self, node, name=None):
        self.node = node
        self._name = str(name)
        self._children = []

    def name(self):
        return self._name

    def attributes(self):
        return { unicode(k): unicode(v) for k, v in self.children() }

    def children(self):
        if self._children:
            return self._children

        if isinstance(self.node, list):
            children = []
            index    = 0

            for item in self.node:
                children.append((index, item))
                index += 1

            self._children = children

            return children

        if isinstance(self.node, dict):
            children = []

            for k in self.node:
                children.append((k, self.node[k]))

            self._children = children

            return children

        return [] # The node is not iterable.

    def value(self):
        return self.node

    def iterable(self):
        return len(self.children()) > 0

    def is_element(self):
        return True

    def is_comment(self):
        return False

    def is_data(self):
        return not self.iterable()

    @staticmethod
    def initialize_children(kotoba_node):
        for name, original_child_node in kotoba_node._node.children():
            node_wrapper = kotoba_node._node.__class__(original_child_node, name) # Instantiate the same node driver.
            child_node   = kotoba_node.__class__(node_wrapper, kotoba_node.level() + 1)

            child_node.parent(kotoba_node)
            kotoba_node.adjacents().append(child_node)

            if child_node.is_element():
                kotoba_node._children.append(child_node)

    @staticmethod
    def retrieve_data(kotoba_node):
        return kotoba_node.original_value()

class XMLDriver(IDriver):
    def __init__(self, node):
        self.node = node

    def name(self):
        return self.node.nodeName

    def attributes(self):
        keys = self.node.attributes.keys()

        return {
            k: self.node.attributes[k].value
            for k in keys
        }

    def children(self):
        for cnode in self.node.childNodes:
            yield cnode

    def value(self):
        return self.node.nodeValue

    def kind(self):
        return self.node.nodeType

    def is_element(self):
        return self.kind() == self.node.ELEMENT_NODE

    def is_comment(self):
        return self.kind() == self.node.COMMENT_NODE

    def is_data(self):
        return self.kind() in [self.node.CDATA_SECTION_NODE, self.node.TEXT_NODE]

    @staticmethod
    def initialize_children(kotoba_node):
        for original_child_node in kotoba_node._node.children():
            node_wrapper = kotoba_node._node.__class__(original_child_node) # Instantiate the same node driver.
            child_node   = kotoba_node.__class__(node_wrapper, kotoba_node.level() + 1)

            if (child_node.is_data() and not child_node.original_value()) or child_node.is_comment():
                continue

            child_node.parent(kotoba_node)
            kotoba_node.adjacents().append(child_node)

            if child_node.is_element():
                kotoba_node._children.append(child_node)

    @staticmethod
    def retrieve_data(kotoba_node):
        data_list = []

        for child in kotoba_node.children(None, True):
            if child.is_data():
                data = child.original_value()

                data_list.append(data)

                continue

            data_list.append(child.data())

        return ''.join(data_list)



from re import compile, sub, split

from .common    import node_debug_message, is_string
from .exception import LexicalError
from .graph     import Vertex as BaseVertex

class PathType(object):
    registered   = [None, '>', '+', '~']
    descendants  = 0
    children     = 1
    any_siblings = 3
    immediate_siblings = 2

class Selector(BaseVertex):
    wildcards = ['','*']

    def __init__(self, parsed_selector_block, _kind=PathType.descendants):
        super(self.__class__, self).__init__()

        self._block = parsed_selector_block

        self._name = None
        self._attributes     = None
        self._pseudo_classes = None
        self._kind = _kind
        self._next = None

        if is_string(self._block):
            self._name  = self._block
            self._block = None

    def name(self):
        if not self._name and self._block:
            self._name = self._block.name

        return self._name

    def attributes(self):
        if not self._attributes:
            self._attributes = []

            if self._block:
                for raw_attribute in self._block.attributes:
                    attribute = Attribute.parse(raw_attribute)

                    self._attributes.append(attribute)

        return self._attributes

    def kind(self, _kind=None):
        if _kind and not self._kind:
            self._kind = _kind

        return self._kind

    def next(self, next=None):
        if next and not self._next:
            self._next = next

        return self._next

    def match(self, vertex):
        node_debug_message(vertex, '[SELECTOR/MATCHING] %s --> %s' % (self.name(), vertex.name()))
        try:
            if not self.name() in self.wildcards:
                assert self.name() == '*' or self.name() == vertex.name()

            for attribute in self.attributes():
                assert attribute.match(vertex)
        except AssertionError:
            return False

        return True

    def attribute(self, name):
        """ Get the attribute by name

            .. versionadded:: 3.1
        """
        attributes = self.attributes()

        for attribute in attributes:
            if attribute.name() == name:
                return attribute.value()

        return None

    def __str__(self):
        return u'SELECTOR %s' % self.name()

class Attribute(object):
    _re_syntax   = compile(r'^\[(?P<name>[^=~\^\$\*\|]+)(?P<operator>[~\^\$\*\|]?=?)(?P<value>.*)\]$')
    _re_operator = compile(r'^[~\^\$\*\|]?=$')

    @staticmethod
    def parse(raw_attribute):
        matches = Attribute._re_syntax.search(raw_attribute)

        if not matches:
            raise LexicalError('The give raw attribute is malform as well as the selector.')

        matches = matches.groupdict()

        if matches['operator']:
            if not Attribute._re_operator.search(matches['operator']):
                raise LexicalError('The operator of the attribute is not recognized.')

            if not matches['value']:
                raise LexicalError('Where is the value?')

            if matches['value'][0] == '"':
                matches['value'] = matches['value'][1:]

            if matches['value'][-1] == '"':
                matches['value'] = matches['value'][:-1]

        return Attribute(**matches)

    def __init__(self, name, operator, value):
        self._name     = name
        self._operator = operator
        self._value    = value

    def name(self):
        return self._name

    def operator(self):
        return self._operator

    def value(self):
        return self._value

    def match(self, vertex):
        try:
            attribute_value = vertex.attribute(self.name())

            assert attribute_value is not None

            if self.operator() == '=':
                assert self.value() == attribute_value

        except AssertionError:
            return False

        return True

    def __str__(self):
        return u'SELECTOR:ATTRIBUTE %s %s %s' % (self.name(), self.operator(), self.value())



import logging
import sys
from . import misc

DEFAULT_LOG_LEVEL = logging.DEBUG

class LoggerRepository(object):
    _cache = {}

    @staticmethod
    def get(name, level=logging.DEBUG, show_time=True):
        if name in LoggerRepository._cache:
            return LoggerRepository._cache[name]

        logging_handler = logging.StreamHandler()
        logging_handler.setLevel(level)
        logging_handler.setFormatter(
            logging.Formatter(
                '%(levelname)s %(asctime)s %(name)s: %(message)s'
                if show_time
                else '%(levelname)s %(name)s: %(message)s',
                datefmt='%Y.%m.%d %H:%M:%S %Z'
            )
        )

        logger = logging.getLogger(name)
        logger.addHandler(logging_handler)
        logger.setLevel(level)

        LoggerRepository._cache[name] = logger

        return logger

def node_debug_message(node, message, ignore_indentation=False):
    if not misc.debug_mode:
        return

    LoggerRepository.get('Node', DEFAULT_LOG_LEVEL).debug(
        message if ignore_indentation else '%s%s' % (' ' * node.level() * 2, message)
    )

def is_string(ref):
    try:
        return type(ref) in [unicode, str]
    except NameError:
        return type(ref) is str